;; Generated package description from .el  -*- hugoistano-byte-compile: t -*-
(define-package "hugoista" "0.2.1" "Manage Hugo posts like a barista" '((emacs "24.3") (seq "0")) :commit "22b543c0cb9c74ede7e762a16d44173c08c906ab" :authors '(("Alexander Adolf" . "alexander.adolf@condition-alpha.com") ("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Alexander Adolf" . "alexander.adolf@condition-alpha.com") :url "https://codeberg.org/c-alpha/hugoista")
